#!/usr/bin/env python

"""GALFIT.PY - Fit galaxy models to extended objects in an image.

"""

__authors__ = 'David Nidever <dnidever@montana.edu?'
__version__ = '20220924'  # yyyymmdd


import os
import sys
import numpy as np
import scipy
import warnings
from astropy.io import fits
from astropy.table import Table
import astropy.units as u
from scipy.optimize import curve_fit, least_squares
from scipy.interpolate import interp1d
from scipy import sparse
#from astropy.nddata import CCDData as CCD,StdDevUncertainty
from dlnpyutils import utils as dln, bindata
import copy
import logging
import time
import matplotlib
import sep
from photutils.aperture import CircularAnnulus
from astropy.stats import sigma_clipped_stats
from . import leastsquares as lsq
from . import groupfit,utils
from .ccddata import CCDData,BoundingBox
from photutils.psf.groupstars import DAOGroup


def galfit(psf,image,cat,method='qr',fitradius=None,recenter=True,maxiter=10,minpercdiff=0.5,
        reskyiter=2,nofreeze=False,skyfit=True,verbose=False):
    """
    Fit galaxy model to all extended objects in an image.

    To pre-group the stars, add a "group_id" in the input catalog.

    Parameters
    ----------
    psf : PSF object
       PSF object with initial parameters to use.
    image : CCDData object
       Image to use to fit PSF model to stars.
    cat : table
       Catalog with initial amp/x/y values for the stars to use to fit the PSF.
       To pre-group the stars, add a "group_id" in the catalog.
    method : str, optional
       Method to use for solving the non-linear least squares problem: "cholesky",
       "qr", "svd", and "curve_fit".  Default is "cholesky".
    fitradius : float, optional
       The fitting radius in pixels.  By default the PSF FWHM is used.
    recenter : boolean, optional
       Allow the centroids to be fit.  Default is True.
    maxiter : int, optional
       Maximum number of iterations to allow.  Only for methods "qr" or "svd".
       Default is 10.
    minpercdiff : float, optional
       Minimum percent change in the parameters to allow until the solution is
       considered converged and the iteration loop is stopped.  Only for methods
       "qr" and "svd".  Default is 0.5.
    reskyiter : int, optional
       After how many iterations to re-calculate the sky background. Default is 2.
    nofreeze : boolean, optional
       Do not freeze any parameters even if they have converged.  Default is False.
    skyfit : boolean, optional
       Fit a constant sky offset with the stellar parameters.  Default is True.
    verbose : boolean, optional
       Verbose output.

    Returns
    -------
    out : table
       Table of best-fitting parameters for each star.
       id, amp, amp_error, x, x_err, y, y_err, sky
    model : numpy array
       Best-fitting model of the stars and sky background.

    Example
    -------

    outcat,model = galfit(psf,image,cat,groups)

    """

    print = utils.getprintfunc() # Get print function to be used locally, allows for easy logging       
    start = time.time()
    
    # Check input catalog
    for n in ['x','y']:
        if n not in cat.keys():
            raise ValueError('Cat must have x and y columns')

    # Check the method
    method = str(method).lower()    
    if method not in ['cholesky','svd','qr','sparse','htcen','curve_fit']:
        raise ValueError('Only cholesky, svd, qr, sparse, htcen or curve_fit methods currently supported')
        
    nstars = np.array(cat).size
    ny,nx = image.data.shape

    
    # Star index
    starindex = dln.create_index(np.array(cat['group_id']))        
    groups = starindex['value']
    ngroups = len(groups)
    if verbose:
        print(str(ngroups)+' star groups')
        
    # Initialize galaxy catalog
    dt = np.dtype([('id',int),('amp',float),('amp_error',float),('x',float),
                   ('x_error',float),('y',float),('y_error',float),('sky',float),
                   ('flux',float),('flux_error',float),('mag',float),('mag_error',float),
                   ('niter',int),('group_id',int),('ngroup',int),('rms',float),('chisq',float)])
    outcat = np.zeros(nstars,dtype=dt)
    outcat = Table(outcat)
    if 'id' in cat.keys():
        outcat['id'] = cat['id']
    else:
        outcat['id'] = np.arange(nstars)+1
        
    # Group Loop
    #---------------
    resid = image.copy()
    outmodel = CCDData(np.zeros(image.shape),bbox=image.bbox,unit=image.unit)
    outsky = CCDData(np.zeros(image.shape),bbox=image.bbox,unit=image.unit)    
    for g,grp in enumerate(groups):
        ind = starindex['index'][starindex['lo'][g]:starindex['hi'][g]+1]
        nind = len(ind)
        inpcat = cat[ind].copy()
        if 'amp' not in inpcat.columns:
            # Estimate amp from flux and fwhm
            # area under 2D Gaussian is 2*pi*A*sigx*sigy
            if 'fwhm' in inpcat.columns:
                amp = inpcat['flux']/(2*np.pi*(inpcat['fwhm']/2.35)**2)
            else:
                amp = inpcat['flux']/(2*np.pi*(psf.fwhm()/2.35)**2)                
            staramp = np.maximum(amp,0)   # make sure it's positive
            inpcat['amp'] = staramp
        
        if verbose:
            print('-- Group '+str(grp)+'/'+str(len(groups))+' : '+str(nind)+' star(s) --')
        
        # Single Star
        if nind==1:
            inpcat = [inpcat['amp'][0],inpcat['x'][0],inpcat['y'][0]]
            out,model = psf.fit(resid,inpcat,niter=3,verbose=verbose,retfullmodel=True,recenter=recenter)
            model.data -= out['sky']   # remove sky
            outmodel.data[model.bbox.slices] += model.data
            outsky.data[model.bbox.slices] = out['sky']

        # Group
        else:
            bbox = cutoutbbox(image,psf,inpcat)
            out,model,sky = groupfit.fit(psf,resid[bbox.slices],inpcat,method=method,fitradius=fitradius,
                                         recenter=recenter,maxiter=maxiter,minpercdiff=minpercdiff,
                                         reskyiter=reskyiter,nofreeze=nofreeze,verbose=verbose,
                                         skyfit=skyfit,absolute=True)
            outmodel.data[model.bbox.slices] += model.data
            outsky.data[model.bbox.slices] = sky
            
        # Subtract the best model for the group/star
        resid[model.bbox.slices].data -= model.data
        
        # Put in catalog
        cols = ['amp','amp_error','x','x_error','y','y_error',
                'sky','flux','flux_error','mag','mag_error','niter','rms','chisq']
        for c in cols:
            outcat[c][ind] = out[c]
        outcat['group_id'][ind] = grp
        outcat['ngroup'][ind] = nind
        outcat = Table(outcat)
        
    if verbose:
        print('dt = %.2f sec' % (time.time()-start))
    
    return outcat,outmodel,outsky



#!/usr/bin/env python

"""UTILS.PY - Some PSF utility routines

"""

__authors__ = 'David Nidever <dnidever@montana.edu?'
__version__ = '20210915'  # yyyymmdd


import os
import sys
import mmap
import numpy as np
import warnings
from astropy.io import fits
from astropy.table import Table
import logging
import time
from datetime import datetime
from scipy.spatial import cKDTree
from dlnpyutils import utils as dln,ladfit
from . import detection, models, getpsf, allfit, leastsquares as lsq
from .ccddata import CCDData
from . import __version__ as version

try:
    import __builtin__ as builtins # Python 2
except ImportError:
    import builtins # Python 3
        
# Ignore these warnings, it's a bug
warnings.filterwarnings("ignore", message="numpy.dtype size changed")
warnings.filterwarnings("ignore", message="numpy.ufunc size changed")

def getprintfunc(inplogger=None):
    """ Allows you to modify print() locally with a logger."""
    
    # Input logger
    if inplogger is not None:
        return inplogger.info  
    # Check if a global logger is defined
    elif hasattr(builtins,"logger"):
        return builtins.logger.info
    # Return the buildin print function
    else:
        return builtins.print

def splitfilename(filename):
    """ Split filename into directory, base and extensions."""
    fdir = os.path.dirname(filename)
    base = os.path.basename(filename)
    exten = ['.fit','.fits','.fit.gz','.fits.gz','.fit.fz','.fits.fz']
    for e in exten:
        if base[-len(e):]==e:
            base = base[0:-len(e)]
            ext = e
            break
    return (fdir,base,ext)


def poly2d(xdata,*pars):
    """ model of 2D linear polynomial."""
    x = xdata[0]
    y = xdata[1]
    return pars[0]+pars[1]*x+pars[2]*y+pars[3]*x*y

def jacpoly2d(xdata,*pars):
    """ jacobian of 2D linear polynomial."""
    x = xdata[0]
    y = xdata[1]
    nx = len(x)
    # Model
    m = pars[0]+pars[1]*x+pars[2]*y+pars[3]*x*y
    # Jacobian, partical derivatives wrt the parameters
    jac = np.zeros((nx,4),float)
    jac[:,0] = 1    # constant coefficient
    jac[:,1] = x    # x-coefficient
    jac[:,2] = y    # y-coefficient
    jac[:,3] = x*y  # xy-coefficient
    return m,jac
    
def poly2dfit(x,y,data,maxiter=2):
    """ Fit a 2D linear function to data robustly."""
    gd, = np.where(np.isfinite(data))
    xdata = [x[gd],y[gd]]
    initpars = np.zeros(4,float)
    med = np.median(data[gd])
    sig = dln.mad(data[gd])
    gd, = np.where( (np.abs(data-med)<3*sig) & np.isfinite(data))
    initpars[0] = med
    xdata = [x[gd],y[gd]]
    # Do the fit
    pars,perror,cov = lsq.lsq_solve(xdata,data[gd],jacpoly2d,initpars,maxiter=maxiter)
    return pars,perror

def estimatefwhm(objects,verbose=False):
    """ Estimate FWHM using objects."""

    print = getprintfunc() # Get print function to be used locally, allows for easy logging   
    
    # Check that we have all of the columns that we need
    for f in ['mag_auto','magerr_auto','flags','fwhm']:
        if f not in objects.colnames:
            raise ValueError('objects catalog must have mag_auto, magerr_auto, flags and fwhm columns')
    
    # Select good sources
    gdobjects = ((objects['mag_auto']< 50) & (objects['magerr_auto']<0.05) &
                 (objects['flags']==0))
    ngdobjects = np.sum(gdobjects)
    # Not enough good source, remove FLAGS cut
    if (ngdobjects<10):
        gdobjects = ((objects['mag_auto']< 50) & (objects['magerr_auto']<0.05))
        ngdobjects = np.sum(gdobjects)
    # Not enough sources, lower thresholds
    if (ngdobjects<10):
        gdobjects = ((objects['mag_auto']< 50) & (objects['magerr_auto']<0.08))
        ngdobjects = np.sum(gdobjects)
    # Not enough sources, lower thresholds
    if (ngdobjects<10):
        si = np.argsort(objects['magerr_auto'])
        halferr = objects['magerr_auto'][si[len(objects)//2]]
        gdobjects = ((objects['mag_auto']< 50) & (objects['magerr_auto']<halferr))
        ngdobjects = np.sum(gdobjects)                    
    medfwhm = np.median(objects[gdobjects]['fwhm'])
    if verbose:
        print('FWHM = %5.2f pixels (%d sources)' % (medfwhm, ngdobjects))

    return medfwhm

def neighbors(objects,nnei=1):
    """ Find the closest neighbors to a star."""

    # Returns distance and index of closest neighbor
    
    # Use KD-tree
    X = np.vstack((objects['x'].data,objects['y'].data)).T
    kdt = cKDTree(X)
    # Get distance for 2 closest neighbors
    dist, ind = kdt.query(X, k=nnei+1)
    # closest neighbor is always itself, remove it
    dist = dist[:,1:]
    ind = ind[:,1:]
    magdiff = objects['mag_auto'][ind[:,0]]-objects['mag_auto']  # negative means neighbor is brighter
    if nnei==1:
        dist = dist.flatten()
        ind = ind.flatten()
    return dist,ind,magdiff
    
def pickpsfstars(objects,fwhm,nstars=100,logger=None,verbose=False):
    """ Pick PSF stars."""

    print = getprintfunc() # Get print function to be used locally, allows for easy logging   
    
    # -morph cuts
    # -magnitude limit (good S/N but not too bright due to saturation)
    # -no bad pixels in footprint
    # -no close neighbors

    # Use KD-tree to figure out closest neighbors
    neidist,neiind,neimagdiff = neighbors(objects)

    # Select good sources
    si = np.argsort(objects['magerr_auto'])
    halferr = objects['magerr_auto'][si[len(objects)//2]]            
    gdobjects1 = ((objects['mag_auto']< 50) & (objects['magerr_auto']<halferr))
    ngdobjects1 = np.sum(gdobjects1)
    # Bright and faint limit, use 5th and 95th percentile
    minmag,maxmag = np.nanpercentile(objects['mag_auto'][gdobjects1],(5,95))
    # Select stars with
    # -good FWHM values
    # -good clas_star values (unless FWHM too large)
    # -good mag range, bright but not too bright
    # -no flags set
    gdobjects = ((objects['mag_auto']<50) & (objects['magerr_auto']<0.05) & 
                 (objects['fwhm']>0.5*fwhm) & (objects['fwhm']<1.5*fwhm) &
                 (objects['mag_auto']>(minmag+1.0)) & (objects['mag_auto']<(maxmag-0.5)) &
                 (objects['flags']==0) & (neidist>15.0) & (neimagdiff>1.0))
    ngdobjects = np.sum(gdobjects)
    # No candidate, loosen cuts
    if ngdobjects<50:
        if verbose:
            print("Too few PSF stars on first try. Loosening cuts")
        gdobjects = ((objects['mag_auto']<50) & (objects['magerr_auto']<0.10) & 
                     (objects['fwhm']>0.2*fwhm) & (objects['fwhm']<1.8*fwhm) &
                     (objects['mag_auto']>(minmag+0.5)) & (objects['mag_auto']<(maxmag-0.5)) &
                     (neidist>10) & (neimagdiff>1.0))
        ngdobjects = np.sum(gdobjects)
    # No candidate, loosen cuts again
    if ngdobjects<50:
        if verbose:
            print("Too few PSF stars on second try. Loosening cuts")
        gdobjects = ((objects['mag_auto']<50) & (objects['magerr_auto']<0.15) & 
                     (objects['fwhm']>0.2*fwhm) & (objects['fwhm']<1.8*fwhm) &
                     (objects['mag_auto']>(minmag+0.5)) & (objects['mag_auto']<(maxmag-0.5)))
        ngdobjects = np.sum(gdobjects)
    # No candidate, loosen cuts again
    if ngdobjects<50:
        if verbose:
            print("Too few PSF stars on second try. Loosening cuts")
        gdobjects = ((objects['mag_auto']<50) & (objects['magerr_auto']<halferr) & 
                     (objects['fwhm']>0.2*fwhm) & (objects['fwhm']<1.8*fwhm) &
                     (objects['mag_auto']>(minmag+0.5)) & (objects['mag_auto']<(maxmag-0.5)))
        ngdobjects = np.sum(gdobjects)
    # No candidates
    if ngdobjects==0:
        raise Exception('No good PSF stars found')
    
    # Candidate PSF stars, use only Nstars, and sort by magnitude
    si = np.argsort(objects[gdobjects]['mag_auto'])
    psfobjects = objects[gdobjects][si]
    if ngdobjects>nstars: psfobjects=psfobjects[0:nstars]
    if verbose:
        print(str(len(psfobjects))+" PSF stars found")
        
    return psfobjects

def refresh_mmap(hdulist):
    """
    Close and refresh an hdulist's memory map to free up virtual memory
    """

    if hasattr(hdulist._file,'_mmap')==False:
        return
    
    MEMMAP_MODES = {'readonly': mmap.ACCESS_COPY,
                'copyonwrite': mmap.ACCESS_COPY,
                'update': mmap.ACCESS_WRITE,
                'append': mmap.ACCESS_COPY,
                'denywrite': mmap.ACCESS_READ}
    access_mode = MEMMAP_MODES[hdulist._file.mode]
    hdulist._file._mmap = mmap.mmap(hdulist._file._file.fileno(), 0,
                                    access=access_mode,
                                    offset=0)
    # Loop over hdus and delete the data in memory
    for h in hdulist:
        if hasattr(h,'data'):
            del h.data

def saveoutput(filename,outfile,out,model,sky,psf):
    """
    Save Prometheus output to a file.

    Parameters
    ----------
    filename : str
       Filename of the original image file.
    outfile : str
       Filename for the output.
    cat : table
       The output table of best-fit PSF values for all of the sources.
    model : CCDData object
       The best-fitting model for the stars (without sky).
    sky : CCDData object
       The background sky image used for the image.
    psf : PSF object
       The best-fitting PSF model.
    
    Returns
    -------
    The output is saved to a FITS file.

    Example
    -------

    saveoutput(filename,outfile,out,model,sky,psf)

    """
    
    if os.path.exists(outfile): os.remove(outfile)
    hdulist = fits.HDUList()
    hdulist.append(fits.table_to_hdu(out))  # table    
    hdulist[0].header['COMMENT']='Prometheus version '+str(version)
    hdulist[0].header['COMMENT']='Date '+datetime.now().ctime()
    hdulist[0].header['COMMENT']='File '+filename
    hdulist[0].header['COMMENT']='HDU#0 : Header Only'
    hdulist[0].header['COMMENT']='HDU#1 : Source catalog'
    hdulist[0].header['COMMENT']='HDU#2 : Model image'
    hdulist[0].header['COMMENT']='HDU#3 : Sky model image'
    hdulist[0].header['COMMENT']='HDU#4 : PSF model'
    hdulist[1].header['COMMENT'] = 'Prometheus source catalog'
    hdulist.append(model.tohdu())  # model
    hdulist[2].header['COMMENT'] = 'Prometheus model image'
    hdulist.append(sky.tohdu())    # sky
    hdulist[3].header['COMMENT'] = 'Prometheus sky image'
    psfhdu = psf.tohdu()
    # psf, could be 2 HDUs
    if isinstance(psfhdu,fits.HDUList):
        for h in psfhdu:
            hdulist.append(h)
        hdulist[4].header['COMMENT'] = 'Prometheus PSF model'
        hdulist[5].header['COMMENT'] = 'Prometheus PSF model lookup table'        
    else:
        hdulist.append(psfhdu)     # psf
        hdulist[4].header['COMMENT'] = 'Prometheus PSF model'
    hdulist.writeto(outfile,overwrite=True)
    hdulist.close()
    



# Fit stars in multiple exposures simultaneously ALLFRAME-style

import os
import numpy as np
from scipy import linalg
from scipy import sparse
from . import leastsquares as lsq, utils

# How to do ALLFRAME-like processing
# -Let's say we have 1 star in 20 exposures and there are 10 pixels in each exposure that contribute
#   to the footprint and that we are fitting.
# -The position of the star in each exposure will be set by the global position and proper motion of the
#   star and is NOT something that is fit at the exposure level
# -At the exposure level we are just fitting the flux.  this is more important when there are multiple,
#    overlapping stars and we need to solve for the fluxes simultaneously.
# -For each exposure, calculate the residuals (data - best model) in the 10 pixels, also calculate the
#   jacobian, i.e the derivatives with respect to the three parameters (x position, y position and flux)
#   for the 10 pixels, [Npix,Npar] or [10,3]
# -With our 20 exposures we now have 20 sets of residuals (10 pixels each), and 20 sets of jacobians
#   (10x3 each).
# -Our goal is to solve for the star position (at some reference time) and proper motion in both
#   X and Y (or RA and DEC).  We will need to solve for the fluxes in each exposure as well.
#   We have 20x3 = 60 exposure-lvel parameters, and we want to solve 2 positions + 2 proper motions +
#   20 exposures = 24 parameters
#   How do we accomplish that?
# -We need a new equation with residuals and jabian for the "global" problem, where our parameters are
#   the 4 positions+proper motion and 20 fluxes.
# -For the residuals it's pretty straighforward.  We want to solve all of the pixels in all the exposures
#   simultaneously.  The pixels are independent, so we just concatenate all of the residuals, giving
#   us 20 x 10 = 200 pixels values.
# -For the jacobian we want the derivative of each pixel with respect to X0, Y0, mu_x, mu_y and the
#   fluxes.  We should be able to get this information from the exposure-level jacobians.
#   The flux part is pretty easy.  We just take the flux column of the relevant exposures's jacobian
#   matrix for the 10 pixels of that exposure and rest of the pixels get zero.
# -The more complicated bit is the derivatives with respect to X0/Y0/mu_x/mu_y.
#   But after some thought this is actually also pretty straightforwad.
#   If I shift X0 slightly by +dx, then the position in each exposure will also be shifted by +dx.
#   So the derivative with respect to X0 is the same as the derivative with respect to the central
#   X position in each exposure.  Therefore, the X0 column of the global jacobian is just the
#   concatenation of the x position columns of the exposure-level jacobians, each one contributing 10
#   values. We end up with 10 pixels x 20 exposures = 200 non-zero values in the X0 column.
# -The mu_x is a little bit more complicated, but not much more.  Here the time lapse with respect to
#   the reference epoch (when X0 and Y0 are defined), is important.  We want to know, if we shift
#   mu_x by a small dmu_x value how does that affect the model for each exposure.  The time lapse is
#   just a multiplicative factor here.  The positional shift for exposure i for a small mu_x shift
#   of dmu_x is just dx_i = t_i * dmu_x.  Since we know what the derivative is for each model with
#   respect to the x position in each exposure, all we need to convert that into a derivative with
#   respect to mu_x is the time factor (t_i).
#   The derivative d/dx_i goes to d/d(t_i * mu_x) = 1/t_i * d/dmu_x or
#   d/dmu_x = t_i * d/dx_i
#   Effectively the exposure-level x position derivative gets simply multiplied by the time lapse
#   (t_i) for that exposure.
#   We populate the mu_x column by again concatenating the exposure-level x position jacobian values,
#   appropriately multipled by their exposure's time laps value.
# -We are left with a new "global" problem/equation with 24 parameters and 200 pixels.  The global
#   jacobian matrix (24 parameters x 200 pixels) is mostly empty.  The four positional columns
#   (X0, Y0, mu_x, mu_y) are fully filled with non-zero values, but the other 20 columns for the
#   exposure-level fluxes are mostly empty/zero except for the 10 pixels for that exposure.
# -We could directly solve this with a simple solver of Ax = b using jacobians.  But since most of
#   the jacobian is empty and "banded", there probably are faster techniques.  I'll have to look
#   at Stetson's ALLFRAME paper to see what he does.  I'm guessing this is the matrix manipulation
#   that you mentioned on the call.
# -I wonder if this could be solved iteratively by first solving for the 4 positional parameters,
#   and then solving the fluxes separately (one flux at a time with its 10 pixels)?  I'm not entirely sure.


def applytrans(trans,data):
    """
    Apply linear transformation to data (rotation and scale).

     xi = CD1_1 * x + CD1_2 * y
    eta = CD2_1 * x + CD2_2 * y

    Parameters
    ----------
    trans : numpy array
       Transformation matrix [2 x 2].
    data : numpy array
       Data to be transformed. The second dimension should have size 2.

    Returns
    -------
    transdata : numpy array
       Transformed data.

    Example
    -------

    transdata = applytrans(trans,data)

    """

    outdata = data.copy()
    outdata[:,0] = trans[0,0] * data[:,0] + trans[1,0] * data[:,1]
    outdata[:,1] = trans[1,0] * data[:,0] + trans[1,1] * data[:,1]
    return outdata


def starfit(data,reftime=None):
    """
    Perform multfit on multiple exposures of a single star.
    
    Parameters
    ----------
    data : list
       List of information.  One element per exposure.  Each element should be 
        a dictionary that contains the following information:
          JAC:  Jacobian matrix [Npix,Npars], where Npix is the number of pixels being
                  fitted for that star and exposure and Npars is 3 for x position, y position
                  and flux.
          RESID: Residuals of the best model-fit and the data (data-model) [Npix].
          WEIGHT: The weights (i.e., 1/sigma^2) of the pixels.
          META:  Dictionary of meta-data.  This should the exposure timestamp information
                  (either DATE-TIME, JD or MJD). And TRANS, the transformation of the X/Y positions
                  in this exposure and at the position of this star, into the reference frame
                  (normally RA/DEC in arcsec).  This must be a 2 x 2 rotation and scaling matrix
                  (like the standard WCS CD matrix).
    reftime : float, optional
       Timestamp to use for the reference epoch.  By default, the first exposure is used.

    Returns
    -------
    pars : numpy array
       Final, best-fit parameters [4 + Nexposures]:  Xref, Yref, mu_x, mu_y (proper motion), and
        amplitudes for each exposure.  Note, these are the CORRECTION terms for the 4 parameters,
        no the actual values themselves.
    perror: numpy array
       Uncertainties in pars.

    Example
    -------

    pars,perror = starfit(data)

    """

    nexp = len(data)

    # Construct the "global" arrays
    #   JAC : Jacobian matrix [Npix total, 4 + Nexposures] for all of the pixels across all
    #            of the exposures and the parameters Xref, Yref, mu_x, mu_y, and amplitudes
    #            for all Nexp exposures.
    #   RESID : residuals of all pixels of all exposures concatenated.
    #   WEIGHT : weights of all pixels of all exposures concatenated.

    # Reference time
    if reftime is None:
        reftime = data[0]['META']['MJD']
    
    # Initialize the global arrays
    npix = np.sum([len(d['RESID']) for d in data])  # add up total number of pixels
    npar = 4 + nexp
    jac = np.zeros([npix,npar],float)
    resid = np.zeros(npix,float)
    weight = np.zeros(npix,float)

    # Loop over the exposures
    pixcount = 0
    for i in range(nexp):
        data1 = data[i]
        jac1 = data1['JAC']
        resid1 = data1['RESID']
        weight1 = data1['WEIGHT']
        meta1 = data1['META']
        time1 = meta1['MJD']
        trans1 = meta['TRANS']
        npix1 = len(resid1)

        # Delta time, time since reference epoch
        delta_time = time1 - reftime   # MJD, days
        delta_time /= 365.2425         # convert to years
        
        # Concatenate RESID and WEIGHT
        resid[count:count+npix1] = resid1
        weight[count:count+npix1] = weight1
            
        # Translate X/Y into the reference frame
        tjac1 = jac1.copy()        
        transjac2 = applytrans(trans1,jac[:,0:2])  # transform X/Y into reference frame
        tjac1[:,0:2] = transjac2

        # Fill in the Global Jacobian
        # Convert X and Y derivatives from the exposure-level jacobian
        #   to the global Xref, Yref, mu_x, and mu_y parameters
        # Xref and Yref derivatives are the same as the X/Y derivatives
        jac[count:count+npix1,0] = tjac1[:,0]              # Xref derivative
        jac[count:count+npix1,1] = tjac1[:,1]              # Yref derivative
        # mu_x and mu_y derivatives are just X derivatives * delta_time        
        jac[count:count+npix1,2] = tjac1[:,0]*delta_time   # mu_x derivative (arcsec/year)
        jac[count:count+npix1,3] = tjac1[:,1]*delta_time   # mu_y derivative (arcsec/year)

        # Amplitude/flux derivatives
        jac[count:count+npix1,i] = jac1[:,2]      # amplitude
        
        # Increment pixcount
        pixcount += npix1

    # Now solve the matrix

    #par = sparse.linalg.lsqr(A,dy2,atol=1e-4,btol=1e-4)
    # pars = lsq.cholesky_jac_sparse_solve(jac,resid,weight)

    # Solve Jacobian
    #  these are the correction terms, no the final parameters themselves
    pars = lsq.jac_solve(jac,resid,method='qr',weight=weight)
    cov = lsq.jac_covariance(jac,resid,weight)
    perror = np.sqrt(np.diag(cov))

    return pars,perror
    



#!/usr/bin/env python

"""SYNTH.PY - Make synthetic star catalogs and images

"""

__authors__ = 'David Nidever <dnidever@montana.edu?'
__version__ = '20210821'  # yyyymmdd


import os
import sys
import numpy as np
import warnings
from astropy.io import fits
from astropy.table import Table
import astropy.units as u
from scipy.optimize import curve_fit, least_squares
from scipy.interpolate import interp1d
from dlnpyutils import utils as dln, bindata
import copy
import logging
import time
import matplotlib
from astropy.nddata import CCDData as CCD,StdDevUncertainty
from .ccddata import CCDData
from . import models

# Fit a PSF model to multiple stars in an image

def makecat(nstars=1000,ampr=[100.0,1e5],xr=[50.0,950.0],yr=[50.0,950.0]):
    """
    Make synthetic catalog of stars.
    """
    dt = np.dtype([('id',int),('x',float),('y',float),('amp',float)])
    cat = np.zeros(nstars,dtype=dt)
    for i in range(nstars):
        amp = dln.scale(np.random.rand(1)[0],[0.0,1.0],ampr)
        xcen = dln.scale(np.random.rand(1)[0],[0.0,1.0],xr)
        ycen = dln.scale(np.random.rand(1)[0],[0.0,1.0],yr)
        cat['id'][i] = i+1
        cat['x'][i] = xcen
        cat['y'][i] = ycen
        cat['amp'][i] = amp
    return cat

def makeimage(nstars=1000,nx=1024,ny=1024,psf=None,cat=None,noise=True,backgrnd=1000.0):
    """
    Make synthetic image
    """

    if psf is None:
        npix = 51
        pix = np.arange(npix)
        pars = [1.0,npix//2,npix/2,3.5,3.6,np.deg2rad(30.)] 
        psf = models.PSFGaussian(pars[3:])   
    fwhm = psf.fwhm()
    npsfpix = psf.npix
    if cat is None:
        cat = makecat(nstars=nstars,xr=[npsfpix/2.0+10,nx-npsfpix/2.0-10],
                      yr=[npsfpix/2.0+10,ny-npsfpix/2.0-10])
    else:
        nstars = len(cat)

    # python images are (Y,X)
    im = np.zeros((ny,nx),float)+backgrnd
    for i in range(nstars):
        amp = cat['amp'][i]
        xcen = cat['x'][i]
        ycen = cat['y'][i]
        bbox = psf.starbbox((xcen,ycen),(ny,nx),npsfpix//2)
        im2 = psf(pars=[amp,xcen,ycen],bbox=bbox)
        im[bbox.slices] += im2
    err = np.maximum(np.sqrt(im),1)
    if noise:
        im += err*np.random.randn(*im.shape)

    image = CCDData(im,err,unit='adu')
    
    return image



#!/usr/bin/env python

"""DETECTION.PY - Detection algorithms

"""

__authors__ = 'David Nidever <dnidever@montana.edu?'
__version__ = '20210912'  # yyyymmdd


import os
import sys
import numpy as np
import warnings
from astropy.io import fits
from astropy.table import Table
import astropy.units as u
from scipy.optimize import curve_fit, least_squares
from scipy.interpolate import interp1d
from dlnpyutils import utils as dln, bindata
import copy
import logging
import time
import matplotlib
from .ccddata import BoundingBox,CCDData
import matplotlib.pyplot as plt
from matplotlib.patches import Ellipse
import sep
from photutils.detection import DAOStarFinder,IRAFStarFinder
from skimage.feature import peak_local_max
    
# A bunch of the Gaussian2D and Moffat2D code comes from astropy's modeling module
# https://docs.astropy.org/en/stable/_modules/astropy/modeling/functional_models.html

# Maybe x0/y0 should NOT be part of the parameters, and
# x/y should actually just be dx/dy (relative to x0/y0)

def plotobj(image,objects):
    """ Plot objects on top of image."""


    # plot background-subtracted image
    fig, ax = plt.subplots()
    m, s = np.mean(image.data), np.std(image.data)
    im = ax.imshow(image.data, interpolation='nearest', cmap='gray',
                   vmin=m-s, vmax=m+s, origin='lower')

    # plot an ellipse for each object
    for i in range(len(objects)):
        if 'a' in objects.colnames:        
            e = Ellipse(xy=(objects['x'][i], objects['y'][i]),
                        width=6*objects['a'][i],
                        height=6*objects['b'][i],
                        angle=objects['theta'][i] * 180. / np.pi)
        elif 'fwhm' in objects.colnames:
            e = Ellipse(xy=(objects['x'][i], objects['y'][i]),
                        width=3*objects['fwhm'][i],
                        height=3*objects['fwhm'][i],
                        angle=0.0)
        else:
            e = Ellipse(xy=(objects['x'][i], objects['y'][i]),
                        width=10.0, height=10.0, angle=0.0)
        e.set_facecolor('none')
        e.set_edgecolor('red')
        ax.add_artist(e)

def sepdetect(image,nsigma=1.5,fwhm=3.0,minarea=3,deblend_nthresh=32,
              deblend_cont=0.000015,kernel=None,maskthresh=0.0,
              segmentation_map=False):
    """ 
    Detection with sep.

    Parameters
    ----------
    image : CCDData object
       Image on which to run sep for detection.
    nsigma : float, optional
       Detection level in units of sigma.  Default is 1.5.
    fwhm : float, optional
       Initial guess fwhm to use.  Default is 3.0.
    minarea : int, optional
       Minimum number of pixels for a detection.  Default is 3.
    deblend_ntresh : int, optional
       Deblending number of thresholds.  Default is 32.
    deblend_cont : float, optional
       Deblending contrast.  Default is 0.000015.
    kernel : numpy array, optional
       Matched filter smoothing kernel.  Default is to construct
         it using the fwhm value.
    maskthresh : float, optional
       Mask threshold.  Default is 0.0.
    segmentation_map : bool, optional
       Return segmentation map.  Default is False.

    Returns
    -------
    objects : table
       Table of the detected objects and properties.
    segmap : numpy array
       Segmentation map if segmentation_map=True was input.

    Example
    -------

    objects = sepdetect(image)

    """
    
    # matched filter
    #  by default a 3x3 kernel is used
    #  to turn off use filter_kernel=None
    #  for optimal detection the kernel size should be approximately the PSF
    if fwhm is not None and kernel is None:
        npix = np.round(1.6*fwhm)
        if npix % 2 == 0: npix += 1
        npix = int(npix)
        x = np.arange(npix).astype(float)-npix//2
        kernel = np.exp(-0.5*( x.reshape(-1,1)**2 + x.reshape(1,-2)**2 )/(fwhm/2.35)**2)
        #kernel = np.array([[1., 2., 3., 2., 1.],
        #                   [2., 3., 5., 3., 2.],
        #                   [3., 5., 8., 5., 3.],
        #                   [2., 3., 5., 3., 2.],
        #                   [1., 2., 3., 2., 1.]])
        #objects = sep.extract(data, thresh, filter_kernel=kernel)
        #kernel = np.array([[1,2,1], [2,4,2], [1,2,1]])  # default 3x3 kernel

            
    # Detection with SEP
    #  NOTE! filter_type='matched' for some reason causes ploblems when 2D error array is input

    # Get C-continuous data
    data,error,mask,sky = image.ccont
    out = sep.extract(data-sky, nsigma, filter_kernel=kernel,minarea=minarea,
                      clean=False,mask=mask, err=error,
                      maskthresh=maskthresh,deblend_nthresh=deblend_nthresh,
                      deblend_cont=deblend_cont,filter_type='conv',
                      segmentation_map=segmentation_map)
    if segmentation_map:
        objects,segmap = out
    else:
        objects = out
    nobj = len(objects)
    objects = Table(objects)
    objects['id'] = np.arange(nobj)+1
    objects['fwhm'] = np.sqrt(objects['a']*objects['b'])*2.35
    objects['flag'].name = 'flags'

    # Make sure theta's are between -90 and +90 deg (in radians)
    hi = objects['theta']>0.5*np.pi
    if np.sum(hi)>0:
        objects['theta'][hi] -= np.pi
    lo = objects['theta']<-0.5*np.pi    
    if np.sum(lo)>0:
        objects['theta'][lo] += np.pi
        
    if segmentation_map:
        return objects,segmap
    else:
        return objects
    
def daodetect(image,nsigma=1.5,fwhm=3.0):
    """ 
    Detection with photutils's DAOFinder.

    Parameters
    ----------
    image : CCDData object
       Image on which to run sep for detection.
    nsigma : float, optional
       Detection level in units of sigma.  Default is 1.5.
    fwhm : float, optional
       Initial guess fwhm to use.  Default is 3.0.
 
    Returns
    -------
    objects : table
       Table of the detected objects and properties.

    Example
    -------

    objects = daodetect(image)

    """

    threshold = np.median(image.error)*nsigma
    daofind = DAOStarFinder(fwhm=fwhm, threshold=threshold, sky=0.0)  
    objects = daofind(image.data-image.sky, mask=image.mask)
    # homogenize the columns
    objects['xcentroid'].name = 'x'
    objects['ycentroid'].name = 'y'
    
    return objects
    
def irafdetect(image,nsigma=1.5,fwhm=3.0):
    """
    Detection with photutil's IRAFFinder.

    Parameters
    ----------
    image : CCDData object
       Image on which to run sep for detection.
    nsigma : float, optional
       Detection level in units of sigma.  Default is 1.5.
    fwhm : float, optional
       Initial guess fwhm to use.  Default is 3.0.
 
    Returns
    -------
    objects : table
       Table of the detected objects and properties.

    Example
    -------

    objects = irafdetect(image)

    """
    threshold = np.median(image.error)*nsigma        
    iraffind = IRAFStarFinder(fwhm=fwhm, threshold=threshold, sky=0.0)
    objects = iraffind(image.data-image.sky, mask=image.mask)
    # homogenize the columns
    objects['xcentroid'].name = 'x'
    objects['ycentroid'].name = 'y'

    return objects

def peaks(image,nsigma=1.5,thresh=None):
    """
    Detect peaks in an image.

    Parameters
    ----------
    image : CCDData object
       Image on which to run sep for detection.
    nsigma : float, optional
       Detection level in units of sigma.  Default is 1.5.
    thresh : float, optional
       Absolute threshold for detection.
 
    Returns
    -------
    objects : table
       Table of the detected peaks.

    Example
    -------

    objects = peaks(image)

    """

    # Comparison between image_max and im to find the coordinates of local maxima
    data = image.data-image.sky
    err = image.error
    coordinates = peak_local_max(data, threshold_abs=thresh, min_distance=3)
    xind = coordinates[:,1]
    yind = coordinates[:,0]    
    
    # Check that they are above the error limit
    nthresh = data[yind,xind]/(nsigma*err[yind,xind])
    if image.mask is None:
        gd, = np.where(nthresh >= 1.0)
    else:
        gd, = np.where((nthresh >= 1.0) & (image.mask[yind,xind]==False))
    nobj = len(gd)
    dtype = np.dtype([('id',int),('x',float),('y',float),('nthresh',float)])
    objects = np.zeros(nobj,dtype=dtype)
    objects['id'] = np.arange(nobj)+1
    objects['x'] = xind[gd]
    objects['y'] = yind[gd]
    objects['nthresh'] = nthresh[gd]
    return objects
    
def detect(image,method='sep',nsigma=1.5,fwhm=3.0,minarea=3,deblend_nthresh=32,
           deblend_cont=0.000015,kernel=None,maskthresh=0.0,verbose=False):
    """
    Detection algorithm

    Parameters
    ----------
    image : CCDData object
       The image to detect sources in.
    method : str, optional
       Method to use.  Options are sep, dao, and iraf.  Default is sep.
    nsigma : float, optional
       Detection threshold in number of sigma.  Default is 1.5.
    fwhm : float, optional
       Estimate for PSF full width at half maximum.  Default is 3.0.
    minarea : int, optional
       Minimum area requirement for an object (sep only).  Default is 3.
    deblend_nthresh : int, optional
       Number of deblending thresholds (sep only).  Default is 32.
    deblend_cont : float, optional
       Minimum contrast ratio used for object deblending (sep only).
         Default is 0.000015.  To entirely disable deblending, set to 1.0.
    kernel : numpy array, optional
        Filter kernel used for on-the-fly filtering (used to
        enhance detection). Default is a 3x3 array.
    maskthresh : float, optional
       Threshold for a pixel to be masked (sep only). Default is 0.0.
    verbose : boolean, optional
       Verbose output to the screen.

    Returns
    -------
    objects : astropy Table
       Table of objects with centroids.
    segmap : numpy array
       Segmentation (sep only).

    Example
    -------

    obj = detect(image,nsigma=1.5,fwhm=4.0)

    """

    if isinstance(image,CCDData) is False:
        raise ValueError("Image must be a CCDData object")

    # Detection method
    method = str(method).lower()

    if verbose:
        print('Detection method = '+method)
        print('Nsigma = %5.2f' % nsigma)
        
    # SEP
    if method=='sep':
        return sepdetect(image,nsigma=nsigma,minarea=minarea,fwhm=fwhm,
                         maskthresh=maskthresh,deblend_nthresh=deblend_nthresh,
                         deblend_cont=deblend_cont,kernel=kernel)


    # DAOFinder
    elif method=='dao':
        return daodetect(image,nsigma=nsigma,fwhm=fwhm)
        return objects
        
    # IRAFFinder
    elif method=='iraf':
        return irafdetect(image,nsigma=nsigma,fwhm=fwhm)
        return objects
        
    else:
        raise ValueError('Only sep, dao or iraf methods supported')



import os
import errno
import numpy as np
from dlnpyutils import utils as dln, robust, coords
from astropy.coordinates import SkyCoord
from astropy.io import fits
from astropy.wcs import WCS
from astropy.time import Time
from astropy.table import Table,vstack,hstack
import astropy.units as u
from . import groupfit,allfit,models,leastsquares as lsq
from .ccddata import CCDData

# ALLFRAME-like forced photometry

# also see multifit.py

def getimageinfo(files):
    """
    Gather information about the images.

    Parameters
    ----------
    files : list
       List of image FITS files.  There should be associated _prometheus.fits
        that include the PSF and source catalog.

    Returns
    -------
    iminfo : table
       List of dictionaries with information on the FITS files including the
        header, WCS, PSF, source catalog, but not the image itself.

    Example
    -------

    iminfo = getimageinfo(files)

    """

    # Load images headers and WCS
    iminfo = []
    nfiles = len(files)
    for i in range(nfiles):
        print('Image {:d}  {:s}'.format(i+1,files[i]))
        if os.path.exists(files[i])==False:
            print(files[i],' not found')
            continue
        prfile = files[i].replace('.fits','_prometheus.fits')            
        if os.path.exists(prfile)==False:
            print(prfile,' not found')
            continue
        # Load header
        head1 = fits.getheader(files[i],0)
        wcs1 = WCS(head1)
        # Load soure table and PSF model from prometheus output file
        tab1 = Table.read(prfile,1)
        for c in tab1.colnames: tab1[c].name = c.lower()
        psf1 = models.read(prfile,4)
        dateobs = head1['DATE-OBS']
        jd = Time(dateobs).jd
        exptime = head1.get('exptime')
        filt = head1.get('filter')
        nx = head1['NAXIS1']
        ny = head1['NAXIS2']        
        cencoo = wcs1.pixel_to_world(nx//2,ny//2)
        vra,vdec = wcs1.wcs_pix2world([0,nx-1,nx-1,0],[0,0,ny-1,ny-1],0)        
        iminfo.append({'file':files[i],'residfile':files[i].replace('.fits','_resid.npy'),
                       'header':head1,'dateobs':dateobs,'jd':jd,'exptime':exptime,'filter':filt,
                       'nx':nx,'ny':ny,'wcs':wcs1,'cenra':cencoo.ra.deg,'cendec':cencoo.dec.deg,
                       'vra':vra,'vdec':vdec,'table':tab1,'psf':psf1,'startmeas':-1,'nmeas':-1})

    return iminfo
    
def makemastertab(images,dcr=1.0,mindet=2):
    """
    Make master star list from individual image star catalogs.

    Parameters
    ----------
    images : list
       List of dictionaries containing the WCS and catalog
        information.
    dcr : float, optional
       Cross-matching radius in arcsec.  Default is 1.0 arcsec.
    mindet : int, optional
       Minimum number of detections to be counted in master list.
        Default is 2.

    Returns
    -------
    mastertab : table
       Table of unique sources in the images.

    Example
    -------

    mastertab = mkmastertab(images)

    """

    nimages = len(images)

    objdt = [('objid',int),('ra',float),('dec',float),('amp',float),
             ('flux',float),('nmeas',int)]
    
    # Loop over the images
    for i in range(nimages):
        tab1 = images[i]['table']
        tab1['objid'] = -1
        tab1['ra'].unit = None   # no units
        tab1['dec'].unit = None        
        # First catalog
        if i==0:
            tab1['objid'] = np.arange(len(tab1))+1
            meas = tab1.copy()
            obj = Table(np.zeros(len(meas),dtype=np.dtype(objdt)))
            for c in ['ra','dec']: obj[c] = meas[c]
            obj['nmeas'] = 1
        # 2nd and later catalog, need to crossmatch
        else:
            # Cross-match
            ind1,ind2,dist = coords.xmatch(obj['ra'],obj['dec'],
                                           tab1['ra'],tab1['dec'],dcr,unique=True)
            # Some matches
            if len(ind1)>0:
                tab1['objid'][ind2] = obj['objid'][ind1]
                obj['nmeas'][ind1] += 1
            # Some left, add them to object table
            if len(ind1) < len(tab1):
                leftind = np.arange(len(tab1))
                leftind = np.delete(leftind,ind2)
                tab1['objid'][leftind] = np.arange(len(leftind))+len(obj)+1
                meas = vstack((meas,tab1))
                newobj = Table(np.zeros(len(leftind),dtype=np.dtype(objdt)))
                for c in ['ra','dec']: newobj[c] = tab1[leftind][c]
                newobj['nmeas'] = 1
                obj = vstack((obj,newobj))

    # Get mean ra, dec and flux from the measurements
    measindex = dln.create_index(meas['objid'])
    obj['flux'] = 0.0
    for i in range(len(measindex['value'])):
        objid = measindex['value'][i]
        ind = measindex['index'][measindex['lo'][i]:measindex['hi'][i]+1]
        nind = len(ind)
        obj['ra'][objid-1] = np.mean(meas['ra'][ind])
        obj['dec'][objid-1] = np.mean(meas['dec'][ind])
        obj['amp'][objid-1] = np.mean(meas['psfamp'][ind])
        obj['flux'][objid-1] = np.mean(meas['psfflux'][ind])

    # Impose minimum number of detections
    if mindet is not None:
        gdobj, = np.where(obj['nmeas'] >= mindet)
        if len(gdobj)==0:
            print('No objects passed the minimum number of detections threshold of '+str(mindet))
            return []
        obj = obj[gdobj]

    return obj

def initialize_catalogs(iminfo,mastertab):
    """
    Initialize the object and measurement array.

    Parameters
    ----------
    iminfo : list
       List of information on the image files.
    mastertab : table
       Master list of sources.

    Returns
    -------
    objtab : table
       Catalog of unique objects.
    meastab : table
       Catalog of individual measurements.
    objindex : list
       List of measurement indices for each object.
    iminfo : table
       List of information on the image files.  Now
        overlap information has been added.

    Example
    -------

    objtab,meastab,objindex,iminfo = initialize_catalogs(iminfo,mastertab)

    """

    nimages = len(iminfo)
    
    nobj = len(mastertab)
    objtab = mastertab.copy()
    objtab['ra'].unit = None   # don't want any units
    objtab['dec'].unit = None    
    objtab['objid'] = 0
    objtab['objid'] = np.arange(nobj)+1
    objtab['cenra0'] = objtab['ra'].copy()  # initial coordinates
    objtab['cendec0'] = objtab['dec'].copy()
    objtab['cenra'] = objtab['ra'].copy()
    objtab['cendec'] = objtab['dec'].copy()    
    objtab['cenpmra'] = 0.0
    objtab['cenpmdec'] = 0.0
    objtab['nmeas'] = 0
    objtab['converged'] = False
    objtab['chisq'] = np.nan
    
    # Initial master list coordinates
    coo0 = SkyCoord(ra=objtab['cenra']*u.deg,dec=objtab['cendec']*u.deg,frame='icrs')
    
    # Get overlap and measurement indices for the images
    meascount = 0
    for i in range(nimages):
        isin = coo0.contained_by(iminfo[i]['wcs'])
        iminfo[i]['startmeas'] = meascount
        iminfo[i]['nmeas'] = np.sum(isin)
        meascount += iminfo[i]['nmeas']
        print('Image {:d} - {:d} stars overlap'.format(i+1,np.sum(isin)))
        
    # Initialize the measurement table
    dt = [('objid',int),('objindex',int),('imindex',int),('jd',float),
          ('ra',float),('dec',float),('x',float),('y',float),('amp',float),
          ('flux',float),('fluxerr',float),('damp',float),('dx',float),
          ('dxerr',float),('dy',float),('dyerr',float),('dra',float),
          ('ddec',float),('sky',float),('chisq',float),('converged',bool)]
    nmeas = np.sum([f['nmeas'] for f in iminfo])
    #nmeas = np.sum(iminfo['nmeas'])
    meastab = Table(np.zeros(nmeas,dtype=np.dtype(dt)))
    meascount = 0
    for i in range(nimages):
        contained = coo0.contained_by(iminfo[i]['wcs'])
        isin, = np.where(contained)
        nisin = len(isin)
        objtab['nmeas'][isin] += 1 
        if nisin > 0:
            meastab['objid'][meascount:meascount+nisin] = objtab['objid'][isin]
            meastab['objindex'][meascount:meascount+nisin] = isin
            meastab['imindex'][meascount:meascount+nisin] = i
            meastab['jd'][meascount:meascount+nisin] = iminfo[i]['jd']
            meascount += nisin

    # Create object index into measurement table
    oindex = dln.create_index(meastab['objindex'])
    objindex = nobj*[[]]  # initialize index with empty list for each object
    for i in range(len(oindex['value'])):
        ind = oindex['index'][oindex['lo'][i]:oindex['hi'][i]+1]
        objindex[oindex['value'][i]] = ind    # value is the index in objtab

    return objtab,meastab,objindex,iminfo


def solveone(psf,im,cat,method='qr',bounds=None,fitradius=None,
             recenter=True,absolute=False,verbose=False):
    """
    Fit the flux and offsets in coordinates for one source.
    
    Parameters
    ----------
    psf : PSF object
       The PSF object for the image.
    im : CCDData object
       Image to use for fitting.
    cat : table
       Initial parameters.  Should at minimum have 'x' and 'y'.
    method : str, optional
       Method to use for solving the non-linear least squares problem: "cholesky",
         "qr", "svd", and "curve_fit".  Default is "qr".
    bounds : list, optional
       Input lower and upper bounds/constraints on the fitting parameters (tuple of two
         lists (e.g., ([amp_lo,x_low,y_low],[amp_hi,x_hi,y_hi])).
    fitradius : float, optional
       Fitting radius in pixels.  Default is to use the PSF FWHM.
    recenter : boolean, optional
       Allow the centroids to be fit.  Default is True.
    absolute : boolean, optional
       Input and output coordinates are in "absolute" values using the image bounding box.
         Default is False, everything is relative.
    verbose : boolean, optional
       Verbose output to the screen.  Default is False.

    Returns
    -------
    newamp : float
       New amplitude.
    dx : float
       The offset in x coordinate.
    dy : float
       The offset in y coordinate.

    Example
    -------
    
    newamp,dx,dy = solveone(psf,im,pars)

    """
    
    method = str(method).lower()
  
    # Image offset for absolute X/Y coordinates
    if absolute:
        imx0 = im.bbox.xrange[0]
        imy0 = im.bbox.yrange[0]

    xc = cat['x'][0]
    yc = cat['y'][0]
    if absolute:  # offset
        xc -= imx0
        yc -= imy0
        if bounds is not None:
            bounds[0][1] -= imx0  # lower
            bounds[0][2] -= imy0
            bounds[1][1] -= imx0  # upper
            bounds[1][2] -= imy0
    if fitradius is None:
        fitradius = np.maximum(psf.fwhm(),1)
    bbox = psf.starbbox((xc,yc),im.shape,fitradius)
    X,Y = psf.bbox2xy(bbox)

    # Get subimage of pixels to fit
    # xc/yc might be offset
    flux = im.data[bbox.slices]
    err = im.error[bbox.slices]
    wt = 1.0/np.maximum(err,1)**2  # weights
    skyim = im.sky[bbox.slices]
    xc -= bbox.ixmin  # offset for the subimage
    yc -= bbox.iymin
    X -= bbox.ixmin
    Y -= bbox.iymin
    if bounds is not None:
        bounds[0][1] -= bbox.ixmin  # lower
        bounds[0][2] -= bbox.iymin
        bounds[1][1] -= bbox.ixmin  # upper
        bounds[1][2] -= bbox.iymin
        
    xdata = np.vstack((X.ravel(), Y.ravel()))        
    #sky = np.median(skyim)
    sky = 0.0
    if 'amp' in cat:
        amp = cat['amp']
    else:
        #amp = flux[int(np.round(yc)),int(np.round(xc))]-sky   # python images are (Y,X)
        #amp = np.maximum(amp,1)  # make sure it's not negative
        amp = 1.0
        
    initpar = [amp,xc,yc,sky]
    
    # Make bounds
    if bounds is None:
        bounds = psf.mkbounds(initpar,flux.shape)
    # Not fitting centroids
    if recenter==False:
        bounds[0][1] = initpar[1]-1e-7
        bounds[0][2] = initpar[2]-1e-7
        bounds[1][1] = initpar[1]+1e-7
        bounds[1][2] = initpar[2]+1e-7
    # Never fitting sky
    bounds[0][3] = -1e-7
    bounds[1][3] = 1e-7    
    
    maxsteps = psf.steps(initpar,bounds,star=True)  # maximum steps
    
    # Use Cholesky, QR or SVD to solve linear system of equations
    m,jac = psf.jac(xdata,*initpar,retmodel=True)
    dy = flux.ravel()-m.ravel()
    # Solve Jacobian
    dbeta = lsq.jac_solve(jac,dy,method=method,weight=wt.ravel())
    dbeta[~np.isfinite(dbeta)] = 0.0  # deal with NaNs
    
    # Perform line search
    alpha,new_dbeta = psf.linesearch(xdata,initpar,dbeta,flux,wt,m,jac) #,allpars=allpars)
                    
    # Update parameters
    oldpar = initpar.copy()
    # limit the steps to the maximum step sizes and boundaries
    newpar = psf.newpars(initpar,new_dbeta,bounds,maxsteps)                
    if recenter==False:
        # Allow the flux to fully vary
        newpar = [np.maximum(new_dbeta[0],0),initpar[1],initpar[2],0.0]
    
    # Output values
    newamp = np.maximum(newpar[0], 0)  # amp cannot be negative
    dx = newpar[1]-initpar[1]
    dy = newpar[2]-initpar[2]

    # Model with only the new flux
    newm = psf.model(xdata,*[newamp,initpar[1],initpar[2],0.0])
    newdy = flux.ravel()-newm.ravel()
    chisq = np.sum(newdy**2 * wt.ravel())/len(newdy)
    
    return newamp,dx,dy,chisq


def solve(psf,resid,tab,fitradius=None,recenter=True,verbose=False):
    """
    Solve for the flux and find corrections for x and y.

    Parameters
    ----------
    psf : psf model
       The image PSF model.
    resid : image
       Residual image with initial estimate of star models subtracted.
    tab : table
       Table of stars ot fit.
    fitradius : float, optional
       The fitting radius in pixels.  The default is 0.5*psf.fwhm().
    recenter : boolean, optional
       Allow the centroids to be fit.  Default is True.
    verbose : bool, optional
       Verbose output to the screen.  Default is False.

    Returns
    -------
    out : table
       Table of results

    Example
    -------

    out = solve(psf,resid,meastab)

    """

    if fitradius is None:
        fitradius = 0.5*psf.fwhm()
    
    ntab = len(tab)
    out = tab.copy()
    
    # Loop over the stars
    for i in range(ntab):
        
        # Add the previous best-fit model back in to the image
        if tab['amp'][i] > 0:
            # nocopy=True will change in place
            _ = psf.add(resid.data,tab[i:i+1],nocopy=True)

        # Solve single flux
        newamp,dx,dy,chisq = solveone(psf,resid,tab[i:i+1],fitradius=fitradius,recenter=recenter)

        # the dx/dy numbers are CRAZY LARGE!!!
        
        # Save the results
        out['amp'][i] = newamp
        out['dx'][i] = dx
        out['dy'][i] = dy
        out['chisq'][i] = chisq
        
        # Immediately subtract the new model
        #  npcopy=True will change in place
        _ = psf.sub(resid.data,out[i:i+1],nocopy=True)

        #print(i,newamp,dx,dy)
        
    return out,resid


def update_object(objtab,meastab,objindex,refepoch,fitpm=False):
    """
    Determine the mean coordinates and proper motions of objects:

    Parameters
    ----------
    objtab : table
       Catalog of unique objects.
    meastab : table
       Catalog of individual measurements.  This should contain the
        updated ra/dec and dra/ddec from the last round of fitting
        sources in the images.
    objindex : list
       List of measurement indices for each object.
    refepoch : Time object
       The reference time to use.  Should be an astropy Time object.
    fitpm : boolean, optional
       Fit proper motions as well as central positions.
         Default is False.

    Returns
    -------
    objtab : table
       Catalog of unique objects with cenra/cendec and pmra/pmdec updated.

    Example
    -------

    objtab = update_object(objtab,meastab,objindex)

    """

    # Loop over object and calculate coordinate and proper motion corrections
    nobj = len(objtab)
    for i in range(nobj):
        cosdec = np.cos(np.deg2rad(objtab['cendec'][i]))
        measind = objindex[i]
        meas1 = meastab[measind]
        jd0 = np.min(meas1['jd'])
        jd = meas1['jd']-jd0
        ra = meas1['ra']+meas1['dra']     # both in degrees
        dec = meas1['dec']+meas1['ddec']
        objtab['chisq'][i] = np.mean(meas1['chisq'])
        # Fit proper motions
        if fitpm:
            # Perform linear fit
            # USE ROBUSTED WEIGHTED LINEAR FIT
            racoef = robust.linefit(jd,ra)
            deccoef = robust.linefit(jd,dec)
            # Get coordinate at the reference epoch
            refra = np.polyval(racoef,refepoch.jd-jd0)
            refdec = np.polyval(deccoef,refepoch.jd-jd0)
            # Calculate new proper motion
            pmra = racoef[0] * (3600*1e3)*365.2425      # convert slope from deg/day to mas/yr
            pmra *= cosdec                              # multiply by cos(dec) for true angle
            pmdec = deccoef[0] * (3600*1e3)*365.2425
            # update object cenra, cendec, cenpmra, cenpmdec
            objtab['cenra'][i] = refra
            objtab['cendec'][i] = refdec
            objtab['cenpmra'][i] = pmra
            objtab['cenpmdec'][i] = pmdec
        # Only fit central coordinates
        else:
            # USE ROBUST WEIGHTED MEAN
            refra = np.mean(ra)
            refdec = np.mean(dec)
            objtab['cenra'][i] = refra
            objtab['cendec'][i] = refdec            

    return objtab


def make_magnitudes(meastab,iminfo,zeropoint=25.0):
    """
    Create measurement magnitudes corrected for exposure time.

    Parameters
    ----------
    meastab : table
       Catalog of individual measurements.  This should contain the
        final fluxes.
    iminfo : list
       List of structures with information on the images including
         exposure times and filter names.
    zeropoint : float, optional
       The zero-point to add to the magnitudes.  Default is 25.0.

    Returns
    -------
    meastab : table
       Catalog of individual measurements.  This will contain a new
        "mag" and "mag_error" magnitude columns.

    Example
    -------

    meastab = make_magnitudes(meastab,iminfo)
    
    """

    # Add the new "mag" column
    meastab['mag'] = 0.0
    meastab['mag_error'] = 0.0    
    
    # Loop over the images
    nimages = len(iminfo)
    for i in range(nimages):
        exptime = iminfo[i]['exptime']
        msbeg = iminfo[i]['startmeas']
        msend = msbeg + iminfo[i]['nmeas']
        flux = meastab[msbeg:msend]['flux']
        mag = -2.5*np.log10(np.maximum(flux,1e-10))+zeropoint
        mag = +2.5*np.log10(exptime)
        flux_error = meastab[msbeg:msend]['flux_error']
        mag_error = (2.5/np.log(10))*(flux_error/flux)
        meastab[msbeg:msend]['mag'] = mag
        meastab[msbeg:msend]['mag_error'] = mag_error        

    return meastab

        
def average_photometry(objtab,meastab,objindex,iminfo):
    """
    Average photometry in each filter for unique objects

    Parameters
    ----------
    objtab : table
       Catalog of unique objects.
    meastab : table
       Catalog of individual measurements.  This should contain the
        final fluxes and "mag" from make_magnitudes().
    objindex : list
       List of measurement indices for each object.
    iminfo : list
       List of structures with information on the images including
         exposure times and filter names.

    Returns
    -------
    objtab : table
       Catalog of unique objects with photometry updated.

    Example
    -------

    objtab = average_photometry(objtab,meastab,objindex)

    """

    # Get unique filters
    imfilters = [f['filter'] for f in iminfo]
    ufilters = np.unique(imfilters)
    nufilt = len(ufilters)
    
    # Add average photometry columns to the object table
    for f in ufilters:
        objtab['mag_'+f] = np.nan
        objtab['mag_error_'+f] = np.nan        
        objtab['ndet_'+f] = 0

    # Image loop
    nobj = len(objtab)
    totalwt = np.zeros((nobj,nufilt),float)
    totalfluxwt = np.zeros((nobj,nufilt),float)    
    ndet = np.zeros((nobj,nufilt),int)
    nimages = len(iminfo)
    for i in range(nimages):
        filt = iminfo[i]['filter']
        ufiltind = np.where(ufilters==filt)[0][0]
        # Get the measurements for this image
        msbeg = iminfo[i]['startmeas']
        msend = msbeg + iminfo[i]['nmeas']
        meas1 = meastab[msbeg:msend]
        # Add up the flux for this filter for all coverage objects
        #   meastab has "objid" which we can use as an index into objtab and totalwt/totalfluxwt/ndet
        totalwt[meas1['objid']-1,ufiltind] += 1.0/meas1['mag_error']**2
        totalfluxwt[meas1['objid']-1,ufiltind] += 2.5118864**meas1['mag'] * (1.0/meas1['mag_error']**2)
        ndet[meas1['objid']-1,ufiltind] += 1

    # Now average the photometry in each unique filter
    for i,f in enumerate(ufilters):
        ind, = np.where(ndet[:,i] > 0)
        objtab['ndet_'+f] = ndet[:,i]
        if len(ind)>0:
            newflux = totalfluxwt[ind,i]/totalwt[ind,i]
            newmag = 2.5*np.log10(newflux) 
            newerr = np.sqrt(1.0/totalwt[ind,i])
            objtab['mag_'+f][ind] = newmag
            objtab['mag_error_'+f][ind] = newerr

    return objtab

def check_convergence(objtab,meastab,objindex,last_obj,fitpm=True):
    """
    Check for convergence of individual stars or all stars.

    Parameters
    ----------
    objtab : table
       Catalog of unique objects.
    meastab : table
       Catalog of individual measurements.  This should contain the
        final fluxes and "mag" from make_magnitudes().
    objindex : list
       List of measurement indices for each object.
    last_obj : table
       Object values from the last iteration.
    fitpm : boolean, optional
       Fit proper motions as well as central positions.
         Default is True.

    Results
    -------
    objtab : table
       Catalog of unique objects.
    meastab : table
       Catalog of individual measurements.  This should contain the
        final fluxes and "mag" from make_magnitudes().
    flag : int
       Flag indicating the convergence.

    Example
    -------

    objtab,meastab,flag = check_convergence(objtab,meastab,objindex)

    """

    flag = 0
    
    # Object loop
    nobj = len(objtab)
    for i in range(nobj):
        measind = objindex[i]
        meas1 = meastab[measind]

        # Check of object cenra/cendec/cenpmra/cenpmdec changed
        #  or the measurement dx/dy values
        dra = objtab['cenra'][i]-last_obj['cenra'][i]
        ddec = objtab['cendec'][i]-last_obj['cendec'][i]
        if fitpm:
            dpmra = objtab['cenpmra'][i]-last_obj['cenpmra'][i]
            dpmdec = objtab['cenpmdec'][i]-last_obj['cenpmdec'][i]

        # Measurement differences
        dx = meas1['dx']
        dy = meas1['dy']

        # Check differences in chisq values
            
    import pdb; pdb.set_trace()

    return objtab,meastab,flag

def forced(files,mastertab=None,fitpm=True,refepoch=None,refwcs=None,verbose=True):
    """
    ALLFRAME-like forced photometry.

    Parameters
    ----------
    files : list
       List of image filenames.
    mastertab : table
       Master table of objects.
    fitpm : boolean, optional
       Fit proper motions as well as central positions.
         Default is True.
    refepoch : Time object, optional
       The reference time to use.  Should be an astropy Time object.
         By default, the mean JD of all images is used.

    Returns
    -------
    obj : table
       Table of unique objects and their mean coordinates and
         proper motions.
    meas : table
       Table of individiual measurements.

    Example
    -------

    obj,meas = forced(files)

    """

    # fit proper motions as well
    # don't load all the data at once, only what you need
    #   maybe use memory maps

    # The default behavior of fits.open() is to use memmap=True
    # and only load the data into physical memory when it is accessed
    # e.g. hdu[0].data.
    # after closing the hdu you still need to delete the data i.e.
    # del hdu[0].data, because there's still a memory map open.

    # I created a function in utils.refresh_mmap() that you can give
    # an open HDUList() and it will refresh the mmap and free up the
    # virtual memory.
    
    nfiles = len(files)
    print('Running forced photometry on {:d} images'.format(nfiles))


    # Do NOT load all of the images at once, only the headers and WCS objects
    # load the PSF and object catalog from the _prometheus.fits file
    # run prometheus if the _prometheus.fits file does not exist
    
    # Load images headers and WCS
    print('Loading image headers, WCS, and catalogs')
    iminfo = getimageinfo(files)
    nimages = len(iminfo)
    if nimages==0:
        print('No images to process')
        return
    
    # Load the master star table if necessary
    if mastertab is not None:
        if isinstance(mastertab,str):
            if os.path.exists(mastertab)==False:
                raise FileNotFoundError(errno.ENOENT, os.strerror(errno.ENOENT), mastertab)
            mastertab_filename = mastertab
            mastertab = Table.read(mastertab_filename)
    # No master star table input, make one from the
    #  individual catalogs
    else:
        print('Creating master list')
        mastertab = makemastertab(iminfo)
    nobj = len(mastertab)
    print('Master list has {:d} stars'.format(nobj))
        
    # Make sure we have the necessary columns
    for c in mastertab.colnames: mastertab[c].name = c.lower()
    if 'ra' not in mastertab.colnames or 'dec' not in mastertab.colnames:
        raise ValueError('ra and dec columns must exist in mastertab')

    # Initialize the array of positions and proper motions
    objtab,meastab,objindex,iminfo = initialize_catalogs(iminfo,mastertab)
    
    # Initial master list coordinates
    coo0 = SkyCoord(ra=objtab['cenra']*u.deg,dec=objtab['cendec']*u.deg,frame='icrs')

    # Some stars have zero measurements
    zeromeas, = np.where(objtab['nmeas']==0)
    if len(zeromeas)>0:
        print('{:d} objects have zero measurements'.format(len(zeromeas)))

    # Reference epoch
    if refepoch is not None:
        if isinstance(refepoch,Time)==False:
            raise ValueError('refepoch must be an astropy Time object')
    else:
        # Default reference epoch is mean JD
        refepoch = Time(np.mean([f['jd'] for f in iminfo]),format='jd')
        
    # Iterate until convergence has been reached
    count = 0
    flag = 0
    #lastvalues = np.zeros(len(meastab),dtype=np.dtype([('flux',float),('dx',float),('dy',float)]))
    while (flag==0):

        print('----- Iteration {:d} -----'.format(count+1))

        if count % 5 == 0:
            print('Recomputing and subtracting the sky')

        # Only fit amplitude on first iteration
        if count==0:
            recenter = False
        else:
            recenter = True

        # Save object information
        ldt = [('cenra',float),('cendec',float),('cenpmra',float),
               ('cenpmdec',float),('chisq',float)]
        last_obj = np.zeros(len(objtab),dtype=np.dtype(ldt))
                                                        
            
        # Loop over the images:
        for i in range(nimages):
            wcs = iminfo[i]['wcs']
            psf = iminfo[i]['psf']
            residfile = iminfo[i]['residfile']
            imtime = Time(iminfo[i]['jd'],format='jd')
            
            # Get the objects and measurements that overlap this image
            contained = coo0.contained_by(iminfo[i]['wcs'])
            objtab1 = objtab[contained]
            msbeg = iminfo[i]['startmeas']
            msend = msbeg + iminfo[i]['nmeas']
            meastab1 = meastab[msbeg:msend]
            print('Image {:d}  {:d} stars'.format(i+1,iminfo[i]['nmeas']))
            
            # Calculate x/y position for each object in this image
            # using the current best overall on-the-sky position
            # and proper motion
            # Need to convert celestial values to x/y position in
            # the image using the WCS.
            coo1 = SkyCoord(ra=objtab1['cenra']*u.deg,dec=objtab1['cendec']*u.deg,
                           pm_ra_cosdec=objtab1['cenpmra']*u.mas/u.year,
                           pm_dec=objtab1['cenpmdec']*u.mas/u.year,
                           obstime=refepoch,frame='icrs')
            
            # Use apply_space_motion() method to get coordinates for the
            #  epoch of this image
            newcoo1 = coo1.apply_space_motion(imtime)
            meastab1['ra'] = newcoo1.ra.deg
            meastab1['dec'] = newcoo1.dec.deg            
            
            # Now convert to image X/Y coordinates
            x,y = wcs.world_to_pixel(newcoo1)
            meastab1['x'] = x
            meastab1['y'] = y            

            # Initialize or load the residual image
            if count==0:
                im = CCDData.read(iminfo[i]['file'])
                resid = im.copy()
            else:
                #resid_data = np.load(residfile)
                #resid.data = resid_data
                reid = dln.unpickle(residfile)
                
            # Subtract sky
            if count % 5 == 0:
                if hasattr(resid,'_sky'):
                    resid._sky = None  # force it to be recomputed
                resid.data -= resid.sky

            # only fit measurements that have not converged yet
                
            # Fit the fluxes while holding the positions fixed            
            out,resid = solve(psf,resid,meastab1,recenter=recenter,verbose=verbose)

            # I THINK THIS IS SLOW BECAUSE IT IS DOING ONE STAR AT A TIME AND
            # USING A LOOP IN PYTHON.  Try to jaxify it!
            
            # Convert dx/dy to dra/ddec
            coo2 = wcs.pixel_to_world(meastab1['x']+out['dx'],meastab1['y']+out['dy'])
            dra = coo2.ra.deg - meastab1['ra']
            ddec = coo2.dec.deg - meastab1['dec']
            out['dra'] = dra    # in degrees
            out['ddec'] = ddec  # in degrees
            
            # Save the residual file
            dln.pickle(residfile,resid)
            
            # SOMETHING IS WRONG, I THINK THE PSF MIGHT NOT BE GOOD

            #from dlnpyutils import plotting as pl 
            #import pdb; pdb.set_trace()
            
            # Stuff the information back in
            meastab['damp'][msbeg:msend] = out['amp']-meastab['amp'][msbeg:msend]
            meastab['amp'][msbeg:msend] = out['amp']
            meastab['ra'][msbeg:msend] = out['ra']
            meastab['dec'][msbeg:msend] = out['dec']
            meastab['dx'][msbeg:msend] = out['dx']
            meastab['dy'][msbeg:msend] = out['dy']
            meastab['dra'][msbeg:msend] = out['dra']
            meastab['ddec'][msbeg:msend] = out['ddec']
            meastab['chisq'][msbeg:msend] = out['chisq']            
            
            # allframe operates on the residual map, with the best-fit model subtracted
            
            # allframe derived flux and centroid corrections for each object
            # the flux corrections are applied immediately while the centroid
            # corrections are saved.

            # there are no groups in allframe
            # the least-squares design matrix is completely diagonalized: the
            # incremental brightness and position corrections are derived for
            # each star separately!  this may add a few more iterations for the
            # badly blended stars, but it does *not* affect the accuracy of the
            # final results.

            # Once a star has converged, it's best-fit model is subtracted from
            # the residual map and its parameters are fixed.
            
            # when no further infinitesimal change to a star's parameters
            # produces any reduction in the robust estimate of the mean-square
            # brightness residual inside that star's fitting region, the
            # maximum-likelihood solution has been achieved.
            
            # Calculate dx/dy residuals for each source
            # convert from pixel to work offset using image wcs

            # need to save the residual image and uncertainties
            # that's all we need to solve the least-squares problem.

            
        # Calculate new coordinates and proper motions based on the x/y residuals
        # the accumulated centroid corrections are projected through the individual
        # frames' geometric transformations to the coordinate system of the master list
        # and averaged.  These net corrections are applied to the stars' positions as
        # retained in the master list.

        # Can also make modest corrections to the input geometric transformation
        # equations by evaluating and removing systematic trends in the centroid
        # corrections derived for stars in each input image.

        # Update the central coordinates and proper motions
        print('Updating object coordinates and proper motions')
        objtab = update_object(objtab,meastab,objindex,refepoch,fitpm=fitpm)

        # Check for convergence
        #   individual stars can converge
        if count > 0:
            objtab,meastab,flag = check_convergence(objtab,meastab,objindex,last_obj,fitpm=fitpm)

        # Save object values 
        last_obj['cenra'] = objtab['cenra']
        last_obj['cendec'] = objtab['cendec']
        last_obj['cenpmra'] = objtab['cenpmra']
        last_obj['cenpmdec'] = objtab['cenpmdec']
        last_obj['chisq'] = objtab['chisq']        
        
        # Check for convergence
        count += 1
            
        import pdb; pdb.set_trace()


    # Create magnitudes, corrected for exposure time
    meastab = make_magnitudes(meastab,iminfo)
        
    # If there's filter information, then get average photometry
    # in each band for the objects
    objtab = average_photometry(objtab,meastab,objindex)
    
        
    return objtab,meastab
